from swarms_cloud.api_utils import (
    VariableInterface,
    check_api_key,
    check_request,
    create_error_response,
    get_model_list,
)
from swarms_cloud.auth_with_swarms_cloud import (
    authenticate_user,
    fetch_api_key_info,
    is_token_valid,
)
from swarms_cloud.calculate_pricing import calculate_pricing, count_tokens
from swarms_cloud.func_api_wrapper import SwarmCloud
from swarms_cloud.log_api_request_to_supabase import ModelAPILogEntry, log_to_supabase
from swarms_cloud.schema.openai_protocol import (  # noqa: E501
    ChatCompletionRequest,
    ChatCompletionRequestQos,
    ChatCompletionResponse,
    ChatCompletionResponseChoice,
    ChatCompletionResponseStreamChoice,
    ChatCompletionStreamResponse,
    ChatMessage,
    CompletionRequest,
    CompletionRequestQos,
    CompletionResponse,
    CompletionResponseChoice,
    CompletionResponseStreamChoice,
    CompletionStreamResponse,
    DeltaMessage,
    EmbeddingsRequest,
    EncodeRequest,
    EncodeResponse,
    ErrorResponse,
    GenerateRequest,
    GenerateRequestQos,
    GenerateResponse,
    GenerationConfig,
    ModelCard,
    ModelList,
    ModelPermission,
    UsageInfo,
)
from swarms_cloud.schema.openai_spec import (
    InputOpenAISpec,
    OpenAIAPIWrapper,
    OutputOpenAISpec,
)
from swarms_cloud.sky_api import SkyInterface
from swarms_cloud.utils.api_key_generator import generate_api_key
from swarms_cloud.utils.rate_limiter import rate_limiter

__all__ = [
    "generate_api_key",
    "SwarmCloud",
    "rate_limiter",
    "SkyInterface",
    "ChatCompletionRequest",
    "ChatCompletionRequestQos",
    "ChatCompletionResponse",
    "ChatCompletionResponseChoice",
    "ChatCompletionResponseStreamChoice",
    "ChatCompletionStreamResponse",
    "ChatMessage",
    "CompletionRequest",
    "CompletionRequestQos",
    "CompletionResponse",
    "CompletionResponseChoice",
    "CompletionResponseStreamChoice",
    "CompletionStreamResponse",
    "DeltaMessage",
    "EmbeddingsRequest",
    "EncodeRequest",
    "EncodeResponse",
    "ErrorResponse",
    "GenerateRequest",
    "GenerateRequestQos",
    "GenerateResponse",
    "ModelCard",
    "ModelList",
    "ModelPermission",
    "UsageInfo",
    "GenerationConfig",
    "VariableInterface",
    "check_api_key",
    "get_model_list",
    "create_error_response",
    "check_request",
    "InputOpenAISpec",
    "OutputOpenAISpec",
    "OpenAIAPIWrapper",
    "calculate_pricing",
    "authenticate_user",
    "is_token_valid",
    "count_tokens",
    "ModelAPILogEntry",
    "log_to_supabase",
    "fetch_api_key_info",
]



import os
from dotenv import load_dotenv
from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
from supabase import create_client, Client


# Load environment variables
load_dotenv()

# HTTP Bearer
http_bearer = HTTPBearer()


# Supabase client
supabase_client_init: Client = create_client(
    supabase_url=os.getenv("SUPABASE_URL"),
    supabase_key=os.getenv("SUPABASE_KEY"),
)


def is_token_valid(token: str, supabase: Client = supabase_client_init) -> bool:
    """
    Check if a token is valid by querying the Supabase database.

    Args:
        token (str): The token to be checked.
        supabase (Client): The Supabase client object.

    Returns:
        bool: True if the token is valid, False otherwise.
    """
    try:
        # Query the Supabase database to check if the token exists in the 'keys' column of the 'swarms_cloud_api_key' table
        response = (
            supabase.table("swarms_cloud_api_keys")
            .select("key")
            .filter("key", "eq", token)
            .execute()
        )
        return response["data"] is not None and len(response["data"]) > 0
    except Exception as error:
        print(f"Error checking token validity: {error}")
        return False


def fetch_api_key_info(token: str, supabase: Client = supabase_client_init):
    """
    Fetch the id and user_id of an API key from the Supabase database.

    Args:
        token (str): The API key to be checked.
        supabase (Client): The Supabase client object.

    Returns:
        dict: A dictionary containing the id and user_id if the API key is valid, None otherwise.
    """
    # Query the Supabase database to check if the token exists in the 'keys' column of the 'swarms_cloud_api_key' table
    try:
        response = (
            supabase.table("swarms_cloud_api_keys")
            .select("id", "user_id", "key")
            .filter("key", "eq", token)
            .execute()
        )
        if response["data"] is not None and len(response["data"]) > 0:
            return response["data"][0]  # return the first matching row
        else:
            return None
    except Exception as error:
        print(f"Error fetching API key info: {error}")
        return None


def authenticate_user(
    credentials: HTTPAuthorizationCredentials = Depends(http_bearer),
):
    """
    Authenticates the user based on the provided credentials.

    Args:
        credentials (HTTPAuthorizationCredentials): The HTTP authorization credentials.

    Returns:
        str: The authentication token.

    Raises:
        HTTPException: If the token is invalid.
    """
    token = credentials.credentials
    if not is_token_valid(token, supabase_client_init):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid token. Please authenticate with a valid token at https://swarms.world/dashboard",
            headers={"WWW-Authenticate": "Bearer"},
        )
    return token



from typing import Any, List

from PIL import Image
from transformers import AutoTokenizer, PreTrainedTokenizer


async def count_tokens(
    texts: List[str], tokenizer: PreTrainedTokenizer, model: str
) -> int:
    """
    Counts the total number of tokens in a list of texts using a tokenizer.

    Args:
        texts (List[str]): A list of texts to count tokens from.
        tokenizer (PreTrainedTokenizer): The tokenizer to use for tokenization.
        model (str): The name or path of the pre-trained model to use.

    Returns:
        int: The total number of tokens in the texts.
    """
    try:
        tokenizer = AutoTokenizer.from_pretrained(model)
        total_tokens = 0
        for text in texts:
            tokens = tokenizer.encode(text, add_special_tokens=True)
            total_tokens += len(tokens)
        return total_tokens
    except Exception as e:
        return e


# Function to calculate tokens and pricing
async def calculate_pricing(
    texts: List[str] = None,
    tokenizer: PreTrainedTokenizer = None,
    images: List[str] = None,
    rate_per_million: float = 0.01,
    img_model: Any = None,
    rate_img: float = 0.003,
    return_cost: bool = True,
    return_tokens: bool = False,
) -> float:
    """
    Calculate containtaining for otal number of  texts based on the number of tokens, sentences, words, characters, and paragraphs.

    Args:
        texts (list): A list of texts to calculate pricing for.
        tokenizer (PreTrainedTokenizer): A pre-trained tokenizer object used to tokenize the texts.
        rate_per_million (float, optional): The rate per million tokens used to calculate the cost. Defaults to 0.01.

    Returns:
        tuple: A tuple containing the total number of tokens, sentences, words, characters, paragraphs, and the calculated cost.

    Example usage:
    >>> tokenizer = AutoTokenizer.from_pretrained("gpt2")
    >>> texts = ["This is the first example text.", "This is the second example text."]
    >>> total_tokens, total_sentences, total_words, total_characters, total_paragraphs, cost = calculate_pricing(texts, tokenizer)
    >>> print(f"Total tokens processed: {total_tokens}")
    >>> print(f"Total cost: ${cost:.5f}")

    """
    total_tokens = 0
    total_images_processed = 0
    image_processing_cost = 0

    for text in texts:
        # Tokenize the text and count tokens
        tokens = tokenizer.encode(text, add_special_tokens=True)
        total_tokens += len(tokens)

    if images and img_model:
        for img_path in images:
            # Load the image
            Image.open(img_path)
            # Process the image
            image_processing_cost += rate_img
            total_images_processed += 1

        # Calculate the image processing cost
        total_images_processed + rate_img

    # Calculate total cost with high precision
    cost = (total_tokens / 1_000_000) * rate_per_million
    print(f"Total cost: ${float(cost):.10f}")

    if return_cost and return_tokens:
        return total_tokens, cost

    if return_cost:
        return cost

    if return_tokens:
        return total_tokens

    return cost



import logging

import stripe
from pydantic import BaseModel

stripe.api_key = "your_stripe_api_key"


class StripeInterface(BaseModel):
    customer_id: str
    amount: float
    description: str


def bill_customer(customer_id: str, amount: float, description: str):
    try:
        stripe.Charge.create(
            customer=customer_id,
            amount=amount,  # in cents
            currency="usd",
            description=description,
        )
        logging.info("Payment successful")
    except stripe.error.StripeError as e:
        logging.error(f"Payment failed: {str(e)}")


# Usage:
# bill_customer("429232323", 1000, "1,000 tokens")



from dotenv import load_dotenv
from pydantic import BaseModel
from supabase import create_client, Client
import os

# Load environment variables from .env file
load_dotenv()


class SupabaseLogger:
    """
    A class for logging records to a Supabase table.

    Args:
        table_name (str): The name of the table to log records to.

    Attributes:
        supabase (Client): The Supabase client used for interacting with the Supabase service.
        table_name (str): The name of the table to log records to.

    Raises:
        ValueError: If the Supabase URL or key is not set in environment variables.

    """

    def __init__(
        self, table_name: str, supabase_url: str = None, supabase_key: str = None
    ):
        if not supabase_url:
            supabase_url = os.getenv("SUPABASE_URL")
        if not supabase_key:
            supabase_key = os.getenv("SUPABASE_KEY")
        if not supabase_url or not supabase_key:
            raise ValueError(
                "Supabase URL and key must be set in environment variables or provided as arguments."
            )
        self.supabase: Client = create_client(supabase_url, supabase_key)
        self.table_name = table_name

    def log(self, record: BaseModel):
        """
        Logs a record to the Supabase table.

        Args:
            record (BaseModel): The record to be logged.

        Returns:
            Response: The response from the Supabase service.

        Raises:
            Exception: If there is an error while logging the record.

        """
        data = record.dict()
        response = self.supabase.table(self.table_name).insert(data).execute()
        if response.error:
            raise Exception(f"Failed to log record: {response.error}")
        return response


# # Usage
# class MyModel(BaseModel):
#     name: str
#     age: int

# logger = SupabaseLogger("my_table")

# record = MyModel(name="John Doe", age=30)
# response = logger.log(record)



import base64
import os
from io import BytesIO
from typing import List, Optional, Tuple

import torch
from loguru import logger
from PIL import Image
from transformers import (
    AutoModelForCausalLM,
    BitsAndBytesConfig,
    LlamaTokenizer,
    PreTrainedModel,
    PreTrainedTokenizer,
    TextIteratorStreamer,
)

from swarms_cloud.schema.cog_vlm_schemas import (
    ChatCompletionResponse,
    ChatCompletionResponseStreamChoice,
    ChatMessageInput,
    DeltaMessage,
    ImageUrlContent,
    TextContent,
)

# Environment variables
MODEL_PATH = os.environ.get("COGVLM_MODEL_PATH", "THUDM/cogvlm-chat-hf")
TOKENIZER_PATH = os.environ.get("TOKENIZER_PATH", "lmsys/vicuna-7b-v1.5")
DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
QUANT_ENABLED = os.environ.get("QUANT_ENABLED", True)


class CogVLMModel:
    def __init__(
        self,
        model_name: str = MODEL_PATH,
        tokenizer_name: str = TOKENIZER_PATH,
        quant_enabled: bool = QUANT_ENABLED,
        device: str = DEVICE,
    ):
        # Load the tokenizer and model
        self.tokenizer = LlamaTokenizer.from_pretrained(
            TOKENIZER_PATH, trust_remote_code=True
        )

        if torch.cuda.is_available() and torch.cuda.get_device_capability()[0] >= 8:
            torch_type = torch.bfloat16
        else:
            torch_type = torch.float16

        self.torch_type = torch_type

        print(
            f"========Use torch type as:{torch_type} with device:{DEVICE}========\n\n"
        )

        quantization_config = {
            "load_in_4bit": True,
            "bnb_4bit_use_double_quant": True,
            "bnb_4bit_compute_dtype": torch_type,
        }

        bnb_config = BitsAndBytesConfig(**quantization_config)

        AutoModelForCausalLM.from_pretrained(
            MODEL_PATH,
            trust_remote_code=True,
            torch_dtype=torch_type,
            low_cpu_mem_usage=True,
            quantization_config=bnb_config,
        ).eval()

        # Torch type
        if torch.cuda.is_available() and torch.cuda.get_device_capability()[0] >= 8:
            torch_type = torch.bfloat16
        else:
            torch_type = torch.float16

    def run(self, task: str, *args, **kwargs):
        pass


async def predict(model_id: str, params: dict):
    """
    Handle streaming predictions. It continuously generates responses for a given input stream.
    This is particularly useful for real-time, continuous interactions with the model.
    """

    choice_data = ChatCompletionResponseStreamChoice(
        index=0, delta=DeltaMessage(role="assistant"), finish_reason=None
    )

    chunk = ChatCompletionResponse(
        model=model_id, choices=[choice_data], object="chat.completion.chunk"
    )

    # Log to supabase
    # supabase_logger.log(chunk)

    yield f"{chunk.model_dump_json(exclude_unset=True)}"

    previous_text = ""
    for new_response in generate_stream_cogvlm(model, tokenizer, params):
        decoded_unicode = new_response["text"]
        delta_text = decoded_unicode[len(previous_text) :]
        previous_text = decoded_unicode
        delta = DeltaMessage(
            content=delta_text,
            role="assistant",
        )
        choice_data = ChatCompletionResponseStreamChoice(
            index=0,
            delta=delta,
        )

        chunk = ChatCompletionResponse(
            model=model_id, choices=[choice_data], object="chat.completion.chunk"
        )

        yield f"{chunk.model_dump_json(exclude_unset=True)}"

    choice_data = ChatCompletionResponseStreamChoice(
        index=0,
        delta=DeltaMessage(),
    )

    chunk = ChatCompletionResponse(
        model=model_id, choices=[choice_data], object="chat.completion.chunk"
    )

    yield f"{chunk.model_dump_json(exclude_unset=True)}"


def generate_cogvlm(
    model: PreTrainedModel, tokenizer: PreTrainedTokenizer, params: dict
):
    """
    Generates a response using the CogVLM model. It processes the chat history and image data, if any,
    and then invokes the model to generate a response.
    """

    for response in generate_stream_cogvlm(model, tokenizer, params):
        pass
    return response


def process_history_and_images(
    messages: List[ChatMessageInput],
) -> Tuple[Optional[str], Optional[List[Tuple[str, str]]], Optional[List[Image.Image]]]:
    """
    Process history messages to extract text, identify the last user query,
    and convert base64 encoded image URLs to PIL images.

    Args:
        messages(List[ChatMessageInput]): List of ChatMessageInput objects.
    return: A tuple of three elements:
             - The last user query as a string.
             - Text history formatted as a list of tuples for the model.
             - List of PIL Image objects extracted from the messages.
    """
    formatted_history = []
    image_list = []
    last_user_query = ""

    for i, message in enumerate(messages):
        role = message.role
        content = message.content

        if isinstance(content, list):  # text
            text_content = " ".join(
                item.text for item in content if isinstance(item, TextContent)
            )
        else:
            text_content = content

        if isinstance(content, list):  # image
            for item in content:
                if isinstance(item, ImageUrlContent):
                    image_url = item.image_url.url
                    if image_url.startswith("data:image/jpeg;base64,"):
                        base64_encoded_image = image_url.split(
                            "data:image/jpeg;base64,"
                        )[1]
                        image_data = base64.b64decode(base64_encoded_image)
                        image = Image.open(BytesIO(image_data)).convert("RGB")
                        image_list.append(image)

        if role == "user":
            if i == len(messages) - 1:  # 最后一条用户消息
                last_user_query = text_content
            else:
                formatted_history.append((text_content, ""))
        elif role == "assistant":
            if formatted_history:
                if formatted_history[-1][1] != "":
                    assert (
                        False
                    ), f"the last query is answered. answer again. {formatted_history[-1][0]}, {formatted_history[-1][1]}, {text_content}"
                formatted_history[-1] = (formatted_history[-1][0], text_content)
            else:
                assert False, "assistant reply before user"
        else:
            assert False, f"unrecognized role: {role}"

    return last_user_query, formatted_history, image_list


@torch.inference_mode()
def generate_stream_cogvlm(
    model: PreTrainedModel, tokenizer: PreTrainedTokenizer, params: dict
):
    """
    Generates a stream of responses using the CogVLM model in inference mode.
    It's optimized to handle continuous input-output interactions with the model in a streaming manner.
    """
    messages = params["messages"]
    temperature = float(params.get("temperature", 1.0))
    repetition_penalty = float(params.get("repetition_penalty", 1.0))
    top_p = float(params.get("top_p", 1.0))
    max_new_tokens = int(params.get("max_tokens", 256))
    query, history, image_list = process_history_and_images(messages)

    logger.debug(f"==== request ====\n{query}")

    input_by_model = model.build_conversation_input_ids(
        tokenizer, query=query, history=history, images=[image_list[-1]]
    )
    inputs = {
        "input_ids": input_by_model["input_ids"].unsqueeze(0).to(DEVICE),
        "token_type_ids": input_by_model["token_type_ids"].unsqueeze(0).to(DEVICE),
        "attention_mask": input_by_model["attention_mask"].unsqueeze(0).to(DEVICE),
        "images": [[input_by_model["images"][0].to(DEVICE).to(torch_type)]],
    }
    if "cross_images" in input_by_model and input_by_model["cross_images"]:
        inputs["cross_images"] = [
            [input_by_model["cross_images"][0].to(DEVICE).to(torch_type)]
        ]

    input_echo_len = len(inputs["input_ids"][0])
    streamer = TextIteratorStreamer(
        tokenizer=tokenizer, timeout=60.0, skip_prompt=True, skip_special_tokens=True
    )
    gen_kwargs = {
        "repetition_penalty": repetition_penalty,
        "max_new_tokens": max_new_tokens,
        "do_sample": True if temperature > 1e-5 else False,
        "top_p": top_p if temperature > 1e-5 else 0,
        "streamer": streamer,
    }
    if temperature > 1e-5:
        gen_kwargs["temperature"] = temperature

    total_len = 0
    generated_text = ""
    with torch.no_grad():
        model.generate(**inputs, **gen_kwargs)
        for next_text in streamer:
            generated_text += next_text
            yield {
                "text": generated_text,
                "usage": {
                    "prompt_tokens": input_echo_len,
                    "completion_tokens": total_len - input_echo_len,
                    "total_tokens": total_len,
                },
            }
    ret = {
        "text": generated_text,
        "usage": {
            "prompt_tokens": input_echo_len,
            "completion_tokens": total_len - input_echo_len,
            "total_tokens": total_len,
        },
    }
    yield ret



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['swarms_cloud',
 'swarms_cloud.loggers',
 'swarms_cloud.schema',
 'swarms_cloud.utils']

package_data = \
{'': ['*']}

install_requires = \
['einops',
 'exxa',
 'fastapi==0.110.0',
 'pydantic>2,<3',
 'shortuuid',
 'skypilot',
 'sse-starlette==2.0.0',
 'stripe',
 'swarms',
 'torch',
 'transformers',
 'uvicorn',
 'xformers']

setup_kwargs = {
    'name': 'swarms-cloud',
    'version': '0.2.2',
    'description': 'Swarms Cloud - Pytorch',
    'long_description': '[![Multi-Modality](agorabanner.png)](https://discord.gg/qUtxnK2NMf)\n\n# Swarms Cloud\nInfrastructure for scalable, reliable, and economical Multi-Modal Model API serving and deployment. We\'re using terraform to orchestrate infrastructure, FastAPI to host the models. If you\'re into deploying models for millions of people, join our discord and help contribute.\n\n\n# Install\n`pip install swarms-cloud`\n\n# Examples\n\n### Example Python\n```python\nimport requests\nimport base64\nfrom PIL import Image\nfrom io import BytesIO\n\n\n# Convert image to Base64\ndef image_to_base64(image_path):\n    with Image.open(image_path) as image:\n        buffered = BytesIO()\n        image.save(buffered, format="JPEG")\n        img_str = base64.b64encode(buffered.getvalue()).decode("utf-8")\n    return img_str\n\n\n# Replace \'image.jpg\' with the path to your image\nbase64_image = image_to_base64("images/3897e80dcb0601c0.jpg")\ntext_data = {"type": "text", "text": "Describe what is in the image"}\nimage_data = {\n    "type": "image_url",\n    "image_url": {"url": f"data:image/jpeg;base64,{base64_image}"},\n}\n\n# Construct the request data\nrequest_data = {\n    "model": "cogvlm-chat-17b",\n    "messages": [{"role": "user", "content": [text_data, image_data]}],\n    "temperature": 0.8,\n    "top_p": 0.9,\n    "max_tokens": 1024,\n}\n\n# Specify the URL of your FastAPI application\nurl = "https://api.swarms.world/v1/chat/completions"\n\n# Send the request\nresponse = requests.post(url, json=request_data)\n# Print the response from the server\nprint(response.text)\n```\n\n### Example API Request in Node\n```js\nconst fs = require(\'fs\');\nconst https = require(\'https\');\nconst sharp = require(\'sharp\');\n\n// Convert image to Base64\nasync function imageToBase64(imagePath) {\n    try {\n        const imageBuffer = await sharp(imagePath).jpeg().toBuffer();\n        return imageBuffer.toString(\'base64\');\n    } catch (error) {\n        console.error(\'Error converting image to Base64:\', error);\n    }\n}\n\n// Main function to execute the workflow\nasync function main() {\n    const base64Image = await imageToBase64("images/3897e80dcb0601c0.jpg");\n    const textData = { type: "text", text: "Describe what is in the image" };\n    const imageData = {\n        type: "image_url",\n        image_url: { url: `data:image/jpeg;base64,${base64Image}` },\n    };\n\n    // Construct the request data\n    const requestData = JSON.stringify({\n        model: "cogvlm-chat-17b",\n        messages: [{ role: "user", content: [textData, imageData] }],\n        temperature: 0.8,\n        top_p: 0.9,\n        max_tokens: 1024,\n    });\n\n    const options = {\n        hostname: \'api.swarms.world\',\n        path: \'/v1/chat/completions\',\n        method: \'POST\',\n        headers: {\n            \'Content-Type\': \'application/json\',\n            \'Content-Length\': requestData.length,\n        },\n    };\n\n    const req = https.request(options, (res) => {\n        let responseBody = \'\';\n\n        res.on(\'data\', (chunk) => {\n            responseBody += chunk;\n        });\n\n        res.on(\'end\', () => {\n            console.log(\'Response:\', responseBody);\n        });\n    });\n\n    req.on(\'error\', (error) => {\n        console.error(error);\n    });\n\n    req.write(requestData);\n    req.end();\n}\n\nmain();\n```\n\n### Example API Request in Go\n\n```go\npackage main\n\nimport (\n    "bytes"\n    "encoding/base64"\n    "encoding/json"\n    "fmt"\n    "image"\n    "image/jpeg"\n    _ "image/png" // Register PNG format\n    "io"\n    "net/http"\n    "os"\n)\n\n// imageToBase64 converts an image to a Base64-encoded string.\nfunc imageToBase64(imagePath string) (string, error) {\n    file, err := os.Open(imagePath)\n    if err != nil {\n        return "", err\n    }\n    defer file.Close()\n\n    img, _, err := image.Decode(file)\n    if err != nil {\n        return "", err\n    }\n\n    buf := new(bytes.Buffer)\n    err = jpeg.Encode(buf, img, nil)\n    if err != nil {\n        return "", err\n    }\n\n    return base64.StdEncoding.EncodeToString(buf.Bytes()), nil\n}\n\n// main is the entry point of the program.\nfunc main() {\n    base64Image, err := imageToBase64("images/3897e80dcb0601c0.jpg")\n    if err != nil {\n        fmt.Println("Error converting image to Base64:", err)\n        return\n    }\n\n    requestData := map[string]interface{}{\n        "model": "cogvlm-chat-17b",\n        "messages": []map[string]interface{}{\n            {\n                "role":    "user",\n                "content": []map[string]string{{"type": "text", "text": "Describe what is in the image"}, {"type": "image_url", "image_url": {"url": fmt.Sprintf("data:image/jpeg;base64,%s", base64Image)}}},\n            },\n        },\n        "temperature": 0.8,\n        "top_p":       0.9,\n        "max_tokens":  1024,\n    }\n\n    requestBody, err := json.Marshal(requestData)\n    if err != nil {\n        fmt.Println("Error marshaling request data:", err)\n        return\n    }\n\n    url := "https://api.swarms.world/v1/chat/completions"\n    request, err := http.NewRequest("POST", url, bytes.NewBuffer(requestBody))\n    if err != nil {\n        fmt.Println("Error creating request:", err)\n        return\n    }\n\n    request.Header.Set("Content-Type", "application/json")\n\n    client := &http.Client{}\n    response, err := client.Do(request)\n    if err != nil {\n        fmt.Println("Error sending request:", err)\n        return\n    }\n    defer response.Body.Close()\n\n    responseBody, err := io.ReadAll(response.Body)\n    if err != nil {\n        fmt.Println("Error reading response body:", err)\n        return\n    }\n\n    fmt.Println("Response:", string(responseBody))\n}\n```\n\n\n\n\n\n## Calculate Pricing\n```python\nfrom transformers import AutoTokenizer\nfrom swarms_cloud import calculate_pricing\n\n# Initialize the tokenizer\ntokenizer = AutoTokenizer.from_pretrained("gpt2")\n\n# Define the example texts\ntexts = ["This is the first example text.", "This is the second example text."]\n\n# Calculate pricing and retrieve the results\ntotal_tokens, total_sentences, total_words, total_characters, total_paragraphs, cost = calculate_pricing(texts, tokenizer)\n\n# Print the total tokens processed\nprint(f"Total tokens processed: {total_tokens}")\n\n# Print the total cost\nprint(f"Total cost: ${cost:.5f}")\n```\n\n\n## Generate an API key\n```python\nfrom swarms_cloud.api_key_generator import generate_api_key\n\nout = generate_api_key(prefix="sk", length=30)\n\nprint(out)\n\n```\n\n# Stack\n- Backend: FastAPI\n- Skypilot for container management\n- Stripe for payment tracking\n- Postresql for database\n- TensorRT for inference\n- Docker for cluster management\n- Kubernetes for managing and autoscaling docker containers\n- Terraform\n\n\n# License\nMIT\n',
    'author': 'Kye Gomez',
    'author_email': 'kye@apac.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kyegomez/swarms-cloud',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import inspect
import logging
from typing import Callable, TypeVar, get_type_hints

import uvicorn
from fastapi import FastAPI, HTTPException, Request, Response
from pydantic import create_model

# Logger initialization
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


# Genertic type for return type
T = TypeVar("T")


# Function API Wrapper for functions: [CLASS]
class SwarmCloud:
    """Functional API Wrapper


    Args:
        host (str, optional): Host to run the API on. Defaults to "
        port (int, optional): Port to run the API on. Defaults to 8000.

    Methods:
        add: Add an endpoint to the API
        run: Run the API

    Example:
    >>> api = SwarmCloud()
    >>> @api.add("/endpoint")
    ... def endpoint():
    ...     return "Hello World"
    >>> api.run()


    """

    def __init__(
        self,
        host: str = "0.0.0.0",
        port: int = 8000,
        local_api: bool = False,
        *args,
        **kwargs,
    ):
        self.host = host
        self.port = port
        self.app = FastAPI()
        self.error_handlers = {}

    def add(self, path: str, method: str = "post", *args, **kwargs):
        """Add an endpoint to the API

        Args:
            path (str): path to add the endpoint to (e.g. "/endpoint")
            method (str, optional): post . Defaults to "post".
        """

        def decorator(func: Callable):
            try:
                if method.lower() == "get":
                    endpoint_func = self.app.get(path)
                elif method.lower() == "post":
                    endpoint_func = self.app.post(path)
                elif method.lower == "put":
                    endpoint_func = self.app.put(path)
                elif method.lower == "delete":
                    endpoint_func = self.app.delete(path)
                elif method.lower == "patch":
                    endpoint_func = self.app.patch(path)

                else:
                    raise ValueError(f"Invalid method: {method}")

                @endpoint_func
                async def endpoint_wrapper(*args, **kwargs):
                    try:
                        logger.info(f"Calling method {func.__name__}")
                        result = func(*args, **kwargs)
                        return result
                    except Exception as error:
                        logger.error(f"Error in {func.__name__}: {error}")
                        raise HTTPException(status_code=500, detail=str(error))

                # Register the endpoint with the http method
                endpoint_func = getattr(self.app, method.lower())
                endpoint_func(path)(endpoint_wrapper)
                return func
            except Exception as error:
                logger.info(f"Error in {func.__name__}: {error}")

        return decorator

    def run(self, *args, **kwargs):
        """Run the API"""
        try:
            uvicorn.run(self.app, host=self.host, port=self.port)
        except Exception as error:
            logger.error(f"Error in {self.__class__.__name__}: {error}")

    def __call__(self, *args, **kwargs):
        """Call the run method

        Args:

        """
        try:
            self.run(*args, **kwargs)
        except Exception as error:
            logger.error(f"Error in {self.__class__.__name__}: {error}")

    def add_endpoints(self, endpoints: list):
        """
        Batch addition of multiple endpoints.

        Args:
            endpoints (list): A list of tuples, each containing path, method, and function.
                              Example: [("/path1", "get", function1), ("/path2", "post", function2)]

        """
        for path, method, func in endpoints:
            self.add(path, method)(func)

    def _generate_request_model(self, func: Callable):
        """Generate requests model

        Args:
            func (Callable): function to generate the request model for

        Returns:
            : dynamically generated request model
        """
        # Extract arguments from the function signature
        signature = inspect.signature(func)
        fields = {
            name: (param.annotation, ...)
            for name, param in signature.parameters.items()
        }
        return create_model(f"{func.__name__}Request", **fields)

    def _generate_response_model(self, func: Callable):
        """Generate response model

        Args:
            func (Callable): function to generate the response model for

        Returns:
            : dynamically generated response model
        """
        return_type = get_type_hints(func).get("return")
        return create_model(f"{func.__name__}Response", result=(return_type, ...))

    def add_error_handler(
        self, exception_class: type, handler: Callable[[Request, Exception], Response]
    ):
        """Add an error handler

        Args:
            exception_class (type): exception class to handle
            handler (Callable[[Request, Exception], Response]): handler function
        """
        self.error_handlers[exception_class] = handler

    async def _call_async_func(self, func, **kwargs):
        """Call an async function

        Args:
            func (callable): function to call

        """
        if inspect.iscoroutinefunction(func):
            return await func(**kwargs)
        else:
            return func(**kwargs)



from http import HTTPStatus
from typing import Any, List, Optional

from fastapi import Depends, HTTPException
from fastapi.responses import JSONResponse
from fastapi.security.http import HTTPAuthorizationCredentials, HTTPBearer

from swarms_cloud.schema.openai_protocol import (  # noqa: E501
    ErrorResponse,
)


class VariableInterface:
    """A IO interface maintaining variables."""

    async_engine: Any = None
    api_keys: Optional[List[str]] = None
    qos_engine: Any = None
    request_hosts = []


get_bearer_token = HTTPBearer(auto_error=False)


async def check_api_key(
    auth: Optional[HTTPAuthorizationCredentials] = Depends(get_bearer_token),
) -> str:
    """Check if client provide valid api key.

    Adopted from https://github.com/lm-sys/FastChat/blob/v0.2.35/fastchat/serve/openai_api_server.py#L108-L127
    """  # noqa
    if VariableInterface.api_keys:
        if (
            auth is None
            or (token := auth.credentials) not in VariableInterface.api_keys
        ):
            raise HTTPException(
                status_code=401,
                detail={
                    "error": {
                        "message": "Please request with valid api key!",
                        "type": "invalid_request_error",
                        "param": None,
                        "code": "invalid_api_key",
                    }
                },
            )
        return token
    else:
        # api_keys not set; allow all
        return None


def get_model_list():
    """Available models."""
    # Use the model registry from swarms
    return [
        "Qwen/Qwen-VL-Chat-Int4",
        "Qwen/Qwen-VL-Chat",
        "Qwen/Qwen-VL",
    ]


def create_error_response(status: HTTPStatus, message: str):
    """Create error response according to http status and message.

    Args:
        status (HTTPStatus): HTTP status codes and reason phrases
        message (str): error message
    """
    return JSONResponse(
        ErrorResponse(
            message=message, type="invalid_request_error", code=status.value
        ).model_dump()
    )


async def check_request(request) -> Optional[JSONResponse]:
    """Check if a request is valid."""
    if request.model in get_model_list():
        return
    ret = create_error_response(
        HTTPStatus.NOT_FOUND, f"The model `{request.model}` does not exist."
    )
    return ret
